"""
BenchNIRS
=========
Benchmarking framework for machine learning with fNIRS
"""

import lazy_loader as lazy

from importlib.metadata import version


try:
    __version__ = version("benchnirs")
except Exception:
    __version__ = "dev"

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        'load': ['load_dataset', 'load_homer'],
        'viz': ['epochs_viz'],
        'process': ['process_epochs', 'extract_features'],
        'learn': ['machine_learn', 'deep_learn', 'deep_transfer_learn',
                  'train_final']
    }
)



import numpy as np

from scipy.stats import linregress, kurtosis, skew


def process_epochs(mne_epochs, tmin=0, tmax=None, tslide=None, sort=False,
                   reject_criteria=None):
    """
    Perform processing on epochs including baseline cropping, bad epoch
    removal, label extraction and  unit conversion.

    Parameters
    ----------
    mne_epochs : MNE Epochs object
        MNE epochs of filtered data with associated labels. Subject IDs are
        contained in the ``metadata`` property.

    tmin : float | None
        Start time of selection in seconds. Defaults to ``0`` to crop the
        baseline.

    tmax : float | None
        End time of selection in seconds. Defaults to ``None`` to keep the
        initial end time.

    tslide : float | None
        Size of the sliding window in seconds. Will crop the epochs if ``tmax``
        is not a multiple of ``tslide``. Defaults to ``None`` for no window
        sliding.

    sort : boolean
        Whether to sort channels by type (all HbO, all HbR). Defaults to
        ``False`` for no sorting.

    reject_criteria : list of floats | None
        List of the 2 peak-to-peak rejection thresholds for HbO and HbR
        channels respectively in uM. Defaults to ``None`` for no rejection.

    Returns
    -------
    nirs : array of shape (n_samples, n_channels, n_times)
        Processed NIRS data in uM.

    labels : array of integer
        List of labels, starting at 0 and with ranks in the same order as
        the event codes. For example if the MNE event codes are
        ``[5, 2, 5, 1]``, the labels will be ``[2, 1, 2, 0]``. Please note that
        ``999`` is reserved for unlabelled samples and will be unchanged.

    groups : array of integer
        List of groups, starting at 0 and with ranks in the same order as the
        original subject IDs. For example if the subject IDs are
        ``[5, 2, 5, 1]``, the groups will be ``[2, 1, 2, 0]``.
    """
    # Process epochs
    epochs = mne_epochs.copy()
    epochs.baseline = None
    epochs.crop(tmin=tmin, tmax=tmax)
    if reject_criteria is not None:
        reject = {'hbo': reject_criteria[0]*1e-6,
                  'hbr': reject_criteria[1]*1e-6}
        epochs.drop_bad(reject=reject)

    # Extract data
    id_list = epochs.events[:, 2]
    sorted_id_set = sorted(epochs.event_id.values())
    id_to_label = {
        id: (i if id != 999 else 999) for i, id in enumerate(sorted_id_set)
    }
    labels = list(map(id_to_label.get, id_list))
    labels = np.array(labels, dtype=np.int64)
    event_to_label = {
        key: id_to_label[value] for key, value in epochs.event_id.items()
    }
    print(f'Event-to-label mapping: {event_to_label}')
    try:
        subj_list = epochs.metadata['Subject_ID'].to_numpy()
        sorted_subj_set = sorted(set(subj_list))
        subj_to_group = {subj: i for i, subj in enumerate(sorted_subj_set)}
        groups = list(map(subj_to_group.get, subj_list))
        groups = np.array(groups, dtype=np.int64)
        print(f'Subject-to-group mapping: {subj_to_group}')
    except TypeError:
        groups = None
    nirs_epochs = epochs.pick('fnirs')
    if sort:
        hbo_chs = epochs.copy().pick('hbo').ch_names
        hbr_chs = epochs.copy().pick('hbr').ch_names
        nirs_epochs.reorder_channels(hbo_chs + hbr_chs)
    nirs = nirs_epochs.get_data(copy=True)
    nirs *= 1e6  # convert from M to uM

    # Create sliding window
    if tslide is not None:
        sliding_size = int(tslide * nirs_epochs.info['sfreq'])
        r = nirs.shape[-1] % sliding_size
        if r > 0:
            nirs = nirs[:, :, :-r]  # crop to fit window size
        nirs = nirs.reshape(nirs.shape[0], nirs.shape[1], -1, sliding_size)
        labels = labels.repeat(nirs.shape[2])
        if groups is not None:
            groups = groups.repeat(nirs.shape[2])
        nirs = nirs.swapaxes(1, 2)
        nirs = nirs.reshape(-1, nirs.shape[2], nirs.shape[3])

    print(f'Dataset shape: {nirs.shape}')

    return nirs, labels, groups


def extract_features(nirs, feature_list):
    """
    Perform feature extraction on NIRS data.

    Parameters
    ----------
    nirs : array of shape (n_samples, n_channels, n_times)
        Processed NIRS data.

    feature_list : list of strings
        List of features to extract. The list can include ``'mean'`` for the
        mean along the time axis, ``'std'`` for standard deviation along the
        time axis and ``'slope'`` for the slope of the linear regression along
        the time axis, ``'skew'`` for the skewness along the time axis,
        ``'kurt'`` for the kurtosis along the time axis, ``'ttp'`` for the time
        to peak (requires channels to have been sorted beforehand: all HbO, all
        HbR), ``'peak'`` for the value of the peak (max value for HbO and min
        value for HbR, requires channels to have been sorted beforehand: all
        HbO, all HbR).

    Returns
    -------
    nirs_features : array of shape (n_samples, n_channels, n_features)
        Features extracted from NIRS data.
    """
    nirs_features = []
    for feature in feature_list:
        if feature == 'mean':
            nirs_feature = np.mean(nirs, axis=-1, keepdims=True)
        elif feature == 'std':
            nirs_feature = np.std(nirs, axis=-1, keepdims=True)
        elif feature == 'slope':
            x = range(nirs.shape[-1])
            nirs_feature = []
            for epoch in nirs:
                ep_slopes = []
                for channel in epoch:
                    ep_slopes.append(linregress(x, channel).slope)
                nirs_feature.append(ep_slopes)
            nirs_feature = np.expand_dims(nirs_feature, -1)
        elif feature == 'skew':
            nirs_feature = skew(nirs, axis=-1, keepdims=True)
        elif feature == 'kurt':
            nirs_feature = kurtosis(nirs, axis=-1, keepdims=True)
        elif feature == 'ttp':
            nirs_hbo, nirs_hbr = np.split(nirs, 2, axis=1)
            hbo_feature = np.argmax(nirs_hbo, axis=-1, keepdims=True)
            hbr_feature = np.argmin(nirs_hbr, axis=-1, keepdims=True)
            nirs_feature = np.concatenate((hbo_feature, hbr_feature), axis=1)
        elif feature == 'peak':
            nirs_hbo, nirs_hbr = np.split(nirs, 2, axis=1)
            hbo_feature = np.max(nirs_hbo, axis=-1, keepdims=True)
            hbr_feature = np.min(nirs_hbr, axis=-1, keepdims=True)
            nirs_feature = np.concatenate((hbo_feature, hbr_feature), axis=1)

        nirs_features.append(nirs_feature)

    nirs_features = np.concatenate(nirs_features, axis=-1)

    return nirs_features



import mne
import numpy as np

from mne.viz import plot_epochs_image, plot_compare_evokeds
from scipy.stats import sem


def _standard_error(x):
    """
    Compute confidence interval bands based on standard error.

    Parameters
    ----------
    x : array of shape (n_observations, n_times)
        The signal observations.

    Returns
    -------
    ci : arrays of shape (2, n_times)
        The confidence interval bands.
    """
    lower_band = np.mean(x, axis=0) + sem(x)
    upper_band = np.mean(x, axis=0) - sem(x)
    ci = np.array([lower_band, upper_band])

    return ci


def epochs_viz(mne_epochs, reject_criteria=None):
    """
    Process and visualize epochs. Processing includes baseline cropping and bad
    epoch removal.

    Parameters
    ----------
    mne_epochs : MNE Epochs object
        MNE epochs with associated labels.

    reject_criteria : list of floats | None
        List of the 2 peak-to-peak rejection thresholds for HbO and HbR
        channels respectively in uM. Defaults to ``None`` for no rejection.
    """
    original_log_lvl = mne.set_log_level(verbose=False, return_old_level=True)

    # Process epochs
    epochs = mne_epochs.copy()
    if reject_criteria is not None:
        reject = {'hbo': reject_criteria[0]*1e-6,
                  'hbr': reject_criteria[1]*1e-6}
        epochs.drop_bad(reject=reject)

    # # Plot all epochs
    # epochs.plot(events=epochs.events, scalings='auto', n_epochs=10,
    #             block=True)

    # # Plot power spectral density
    # epochs.compute_psd(fmin=0, fmax=5).plot(average=True)

    # Visualise with average epochs superposed
    conditions = epochs.event_id.keys()
    for condition in conditions:
        plot_epochs_image(epochs[condition].copy().pick('hbo'), combine='mean',
                          title=f'{condition} (HbO)')
        plot_epochs_image(epochs[condition].copy().pick('hbr'), combine='mean',
                          title=f'{condition} (HbR)')

    # Visualise condition comparison
    evoked_dict = {}
    styles_dict = {}
    linestyles = ['solid', 'dashed', 'dotted', 'dashdot', 'dashdotdotted']
    for idx, condition in enumerate(conditions):
        evs_hbo = list(epochs[condition].copy().pick('hbo').iter_evoked())
        evs_hbr = list(epochs[condition].copy().pick('hbr').iter_evoked())
        for ev in evs_hbo:
            ev.rename_channels(lambda x: x[:-4])  # remove ' HbO' from ch_names
        for ev in evs_hbr:
            ev.rename_channels(lambda x: x[:-4])  # remove ' HbR' from ch_names
        evoked_dict[f'{condition}/HbO'] = evs_hbo
        evoked_dict[f'{condition}/HbR'] = evs_hbr
        styles_dict[condition] = {
            'linestyle': linestyles[idx % len(linestyles)]
        }
    color_dict = {'HbO': '#AA3377', 'HbR': 'b'}
    plot_compare_evokeds(evoked_dict, combine='mean', ci=0.95,
                         colors=color_dict, styles=styles_dict,
                         title='Conditions with 0.95 CI')
    plot_compare_evokeds(evoked_dict, combine='mean', ci=_standard_error,
                         colors=color_dict, styles=styles_dict,
                         title='Conditions with standard error')

    # Get only HbO and HbR and remove STIM channel
    nirs_epochs = epochs.pick('fnirs')
    print(nirs_epochs)

    # Get fnirs data as a numpy array
    channels = nirs_epochs.info['ch_names']
    nirs = nirs_epochs.get_data(copy=True)
    print(f'Channels: {channels}')
    print(f'NIRS data shape: {nirs.shape}')

    # Get label list
    labels = nirs_epochs.events[:, 2]
    print(f'Labels shape: {labels.shape}')

    # Get standard deviations
    for label_class in np.unique(labels):
        indices_class = [i for i, x in enumerate(labels) if x == label_class]
        nirs_class = np.take(nirs, indices_class, axis=0)
        std_class = np.mean(np.std(nirs_class, axis=0))
        std_class = round(std_class * 1e6, 3)  # from M to uM
        print('Standard deviation across trials, averaged across channels '
              f'and time [condition {label_class}]: {std_class} uM')

    mne.set_log_level(verbose=original_log_lvl)



import glob
import mne
import nirsimple as ns
import numpy as np
import scipy.io

from mne.channels import combine_channels
from pandas import DataFrame
from pathlib import Path
from sys import stdout
from tqdm import tqdm


PATH_HERFF_2014 = '../../data/dataset_herff_2014/'
PATH_SHIN_2018 = '../../data/dataset_shin_2018/'
PATH_SHIN_2016 = '../../data/dataset_shin_2016/'
PATH_BAK_2019 = '../../data/dataset_bak_2019/'


def _temporal_derivative_distribution_repair(raw):
    """
    Function taken from MNE-Python and modified to work on non-standard
    channels

    # Authors: Robert Luke <mail@robertluke.net> and Frank Fishburn
    # License: BSD-3-Clause

    Apply temporal derivative distribution repair to data.

    Applies temporal derivative distribution repair (TDDR) to data
    (Fishburn et al. 2019). This approach removes baseline shift
    and spike artifacts without the need for any user-supplied parameters.

    Parameters
    ----------
    raw : instance of Raw
        The raw data.

    Returns
    -------
    raw : instance of Raw
         Data with TDDR applied.
    """
    from mne.io import BaseRaw
    from mne.utils import _validate_type
    from mne.io.pick import _picks_to_idx

    raw = raw.copy().load_data()
    _validate_type(raw, BaseRaw, 'raw')

    picks = _picks_to_idx(raw.info, 'fnirs', exclude=[])
    for pick in picks:
        raw._data[pick] = _TDDR(raw._data[pick], raw.info['sfreq'])

    return raw


def _TDDR(signal, sample_rate):
    """
    Function taken from MNE-Python

    # Authors: Robert Luke <mail@robertluke.net> and Frank Fishburn
    # License: BSD-3-Clause
    """
    # This function is the reference implementation for the TDDR algorithm for
    #   motion correction of fNIRS data, as described in:
    #
    #   Fishburn F.A., Ludlum R.S., Vaidya C.J., & Medvedev A.V. (2019).
    #   Temporal Derivative Distribution Repair (TDDR): A motion correction
    #   method for fNIRS. NeuroImage, 184, 171-179.
    #   https://doi.org/10.1016/j.neuroimage.2018.09.025
    #
    # Usage:
    #   signals_corrected = TDDR( signals , sample_rate );
    #
    # Inputs:
    #   signals: A [sample x channel] matrix of uncorrected optical density or
    #            hb data
    #   sample_rate: A scalar reflecting the rate of acquisition in Hz
    #
    # Outputs:
    #   signals_corrected: A [sample x channel] matrix of corrected optical
    #   density data
    from scipy.signal import butter, filtfilt
    signal = np.array(signal)
    if len(signal.shape) != 1:
        for ch in range(signal.shape[1]):
            signal[:, ch] = _TDDR(signal[:, ch], sample_rate)
        return signal

    # Preprocess: Separate high and low frequencies
    filter_cutoff = .5
    filter_order = 3
    Fc = filter_cutoff * 2 / sample_rate
    signal_mean = np.mean(signal)
    signal -= signal_mean
    if Fc < 1:
        fb, fa = butter(filter_order, Fc)
        signal_low = filtfilt(fb, fa, signal, padlen=0)
    else:
        signal_low = signal

    signal_high = signal - signal_low

    # Initialize
    tune = 4.685
    D = np.sqrt(np.finfo(signal.dtype).eps)
    mu = np.inf
    iter = 0

    # Step 1. Compute temporal derivative of the signal
    deriv = np.diff(signal_low)

    # Step 2. Initialize observation weights
    w = np.ones(deriv.shape)

    # Step 3. Iterative estimation of robust weights
    while iter < 50:

        iter = iter + 1
        mu0 = mu

        # Step 3a. Estimate weighted mean
        mu = np.sum(w * deriv) / np.sum(w)

        # Step 3b. Calculate absolute residuals of estimate
        dev = np.abs(deriv - mu)

        # Step 3c. Robust estimate of standard deviation of the residuals
        sigma = 1.4826 * np.median(dev)

        # Step 3d. Scale deviations by standard deviation and tuning parameter
        if sigma == 0:
            break
        r = dev / (sigma * tune)

        # Step 3e. Calculate new weights according to Tukey's biweight function
        w = ((1 - r**2) * (r < 1)) ** 2

        # Step 3f. Terminate if new estimate is within
        # machine-precision of old estimate
        if abs(mu - mu0) < D * max(abs(mu), abs(mu0)):
            break

    # Step 4. Apply robust weights to centered derivative
    new_deriv = w * (deriv - mu)

    # Step 5. Integrate corrected derivative
    signal_low_corrected = np.cumsum(np.insert(new_deriv, 0, 0.0))

    # Postprocess: Center the corrected signal
    signal_low_corrected = signal_low_corrected - np.mean(signal_low_corrected)

    # Postprocess: Merge back with uncorrected high frequency component
    signal_corrected = signal_low_corrected + signal_high + signal_mean

    return signal_corrected


class _DatasetHerff2014NB():
    """
    Loader for n-back data from Herff et al., 2014.
    """

    def __init__(self):
        self.subject_list = ['subject1', 'subject2', 'subject3', 'subject4',
                             'subject5', 'subject6', 'subject7', 'subject8',
                             'subject9', 'subject10']

    def load(self, subject, path=PATH_HERFF_2014):
        data = {}
        data['sfreq'] = 25  # Hz
        data['ch_names'] = ['CH1 HbO', 'CH1 HbR', 'CH2 HbO', 'CH2 HbR',
                            'CH3 HbO', 'CH3 HbR', 'CH4 HbO', 'CH4 HbR',
                            'CH5 HbO', 'CH5 HbR', 'CH6 HbO', 'CH6 HbR',
                            'CH7 HbO', 'CH7 HbR', 'CH8 HbO', 'CH8 HbR']
        data['ch_types'] = ['hbo', 'hbr', 'hbo', 'hbr',
                            'hbo', 'hbr', 'hbo', 'hbr',
                            'hbo', 'hbr', 'hbo', 'hbr',
                            'hbo', 'hbr', 'hbo', 'hbr']
        data['sides'] = {'Right PFC HbO': [0, 2, 4, 6],
                         'Right PFC HbR': [1, 3, 5, 7],
                         'Left PFC HbO': [8, 10, 12, 14],
                         'Left PFC HbR': [9, 11, 13, 15]}
        data['event_id'] = {'1-back': 1, '2-back': 2, '3-back': 3}
        data['tmin'] = -5
        data['tmax'] = 44

        # Load data matrix
        mat = scipy.io.loadmat(f'{path}/{subject}.mat')

        # Get Hb data
        delta_c = mat['data']  # in uM
        delta_c = delta_c * 1e-6  # convert uM to M
        data['delta_c'] = delta_c

        # Get labels
        labels = mat['label']
        labels[labels == -1] = 6  # replace -1 by 6 to avoid Epochs issue
        data['labels'] = labels

        return data


class _DatasetShin2018NB():
    """
    Loader for n-back data from Shin et al., 2018.
    """

    def __init__(self):
        self.subject_list = ['VP001-NIRS', 'VP002-NIRS', 'VP003-NIRS',
                             'VP004-NIRS', 'VP005-NIRS', 'VP006-NIRS',
                             'VP007-NIRS', 'VP008-NIRS', 'VP009-NIRS',
                             'VP010-NIRS', 'VP011-NIRS', 'VP012-NIRS',
                             'VP013-NIRS', 'VP014-NIRS', 'VP015-NIRS',
                             'VP016-NIRS', 'VP017-NIRS', 'VP018-NIRS',
                             'VP019-NIRS', 'VP020-NIRS', 'VP021-NIRS',
                             'VP022-NIRS', 'VP023-NIRS', 'VP024-NIRS',
                             'VP025-NIRS', 'VP026-NIRS']

    def load(self, subject, path=PATH_SHIN_2018):
        data = {}
        data['sides'] = {'Right PFC HbO': [9, 10, 19, 20, 21, 22, 23],
                         'Right PFC HbR': [45, 46, 55, 56, 57, 58, 59],
                         'Left PFC HbO': [0, 1, 2, 3, 4, 5, 6],
                         'Left PFC HbR': [36, 37, 38, 39, 40, 41, 42]}
        data['event_id'] = {'0-back': 1, '2-back': 2, '3-back': 3}
        data['tmin'] = -2
        data['tmax'] = 40

        # Load data matrices
        cnt = scipy.io.loadmat(f'{path}/{subject}/cnt_nback.mat')
        mrk = scipy.io.loadmat(f'{path}/{subject}/mrk_nback.mat')

        # Get sampling frequency (in Hz)
        data['sfreq'] = cnt['cnt_nback']['oxy'][0, 0]['fs'][0, 0][0][0]

        # Get channel info
        ch_oxy = cnt['cnt_nback']['oxy'][0, 0]['clab'][0, 0][0]
        ch_deoxy = cnt['cnt_nback']['deoxy'][0, 0]['clab'][0, 0][0]
        ch_names_oxy = [f'{ch[0]} HbO' for ch in ch_oxy]
        ch_names_deoxy = [f'{ch[0]} HbR' for ch in ch_deoxy]
        ch_types_oxy = ['hbo' for ch in ch_names_oxy]
        ch_types_deoxy = ['hbr' for ch in ch_names_deoxy]
        data['ch_names'] = ch_names_oxy + ch_names_deoxy
        data['ch_types'] = ch_types_oxy + ch_types_deoxy

        # Get Hb data
        delta_oxy = cnt['cnt_nback']['oxy'][0, 0]['x'][0, 0]  # in mM
        delta_deoxy = cnt['cnt_nback']['deoxy'][0, 0]['x'][0, 0]  # in mM
        delta_c = np.hstack((delta_oxy, delta_deoxy))
        delta_c = delta_c * 1e-3  # convert mM to M
        delta_c = np.swapaxes(delta_c, 0, 1)
        data['delta_c'] = delta_c

        # Get labels
        markers_time = mrk['mrk_nback']['time'][0, 0]/1000*data['sfreq']
        markers_time = np.around(markers_time).astype(int)
        y_idx_0back = np.nonzero(mrk['mrk_nback']['y'][0, 0][0])
        y_idx_2back = np.nonzero(mrk['mrk_nback']['y'][0, 0][1])
        y_idx_3back = np.nonzero(mrk['mrk_nback']['y'][0, 0][2])
        time_0back = markers_time[0, y_idx_0back]
        time_2back = markers_time[0, y_idx_2back]
        time_3back = markers_time[0, y_idx_3back]
        labels = np.zeros((1, delta_c.shape[-1]))
        labels[0, time_0back] = 1
        labels[0, time_2back] = 2
        labels[0, time_3back] = 3
        data['labels'] = labels

        return data


class _DatasetShin2018WG():
    """
    Loader for word generation data from Shin et al., 2018.
    """

    def __init__(self):
        self.subject_list = ['VP001-NIRS', 'VP002-NIRS', 'VP003-NIRS',
                             'VP004-NIRS', 'VP005-NIRS', 'VP006-NIRS',
                             'VP007-NIRS', 'VP008-NIRS', 'VP009-NIRS',
                             'VP010-NIRS', 'VP011-NIRS', 'VP012-NIRS',
                             'VP013-NIRS', 'VP014-NIRS', 'VP015-NIRS',
                             'VP016-NIRS', 'VP017-NIRS', 'VP018-NIRS',
                             'VP019-NIRS', 'VP020-NIRS', 'VP021-NIRS',
                             'VP022-NIRS', 'VP023-NIRS', 'VP024-NIRS',
                             'VP025-NIRS', 'VP026-NIRS']

    def load(self, subject, path=PATH_SHIN_2018):
        data = {}
        data['sides'] = {'Right PFC HbO': [9, 10, 19, 20, 21, 22, 23],
                         'Right PFC HbR': [45, 46, 55, 56, 57, 58, 59],
                         'Left PFC HbO': [0, 1, 2, 3, 4, 5, 6],
                         'Left PFC HbR': [36, 37, 38, 39, 40, 41, 42]}
        data['event_id'] = {'baseline': 1, 'word generation': 2}
        data['tmin'] = -2
        data['tmax'] = 10

        # Load data matrices
        cnt = scipy.io.loadmat(f'{path}/{subject}/cnt_wg.mat')
        mrk = scipy.io.loadmat(f'{path}/{subject}/mrk_wg.mat')

        # Get sampling frequency (in Hz)
        data['sfreq'] = cnt['cnt_wg']['oxy'][0, 0]['fs'][0, 0][0][0]

        # Get channel info
        ch_oxy = cnt['cnt_wg']['oxy'][0, 0]['clab'][0, 0][0]
        ch_deoxy = cnt['cnt_wg']['deoxy'][0, 0]['clab'][0, 0][0]
        ch_names_oxy = [f'{ch[0]} HbO' for ch in ch_oxy]
        ch_names_deoxy = [f'{ch[0]} HbR' for ch in ch_deoxy]
        ch_types_oxy = ['hbo' for ch in ch_names_oxy]
        ch_types_deoxy = ['hbr' for ch in ch_names_deoxy]
        data['ch_names'] = ch_names_oxy + ch_names_deoxy
        data['ch_types'] = ch_types_oxy + ch_types_deoxy

        # Get Hb data
        delta_oxy = cnt['cnt_wg']['oxy'][0, 0]['x'][0, 0]  # in mM
        delta_deoxy = cnt['cnt_wg']['deoxy'][0, 0]['x'][0, 0]  # in mM
        delta_c = np.hstack((delta_oxy, delta_deoxy))
        delta_c = delta_c * 1e-3  # convert mM to M
        delta_c = np.swapaxes(delta_c, 0, 1)
        data['delta_c'] = delta_c

        # Get labels
        markers_time = mrk['mrk_wg']['time'][0, 0]/1000*data['sfreq']
        markers_time = np.around(markers_time).astype(int)
        y_idx_wg = np.nonzero(mrk['mrk_wg']['y'][0, 0][0])
        y_idx_bl = np.nonzero(mrk['mrk_wg']['y'][0, 0][1])
        time_wg = markers_time[0, y_idx_wg]
        time_bl = markers_time[0, y_idx_bl]
        labels = np.zeros((1, delta_c.shape[-1]))
        labels[0, time_wg] = 2
        labels[0, time_bl] = 1
        data['labels'] = labels

        return data


class _DatasetShin2016MA():
    """
    Loader for mental arithmetic data from Shin et al., 2016.
    """

    def __init__(self):
        self.subject_list = ['subject 01', 'subject 02', 'subject 03',
                             'subject 04', 'subject 05', 'subject 06',
                             'subject 07', 'subject 08', 'subject 09',
                             'subject 10', 'subject 11', 'subject 12',
                             'subject 13', 'subject 14', 'subject 15',
                             'subject 16', 'subject 17', 'subject 18',
                             'subject 19', 'subject 20', 'subject 21',
                             'subject 22', 'subject 23', 'subject 24',
                             'subject 25', 'subject 26', 'subject 27',
                             'subject 28', 'subject 29']

    def load(self, subject, path=PATH_SHIN_2016):
        data = {}
        data['sides'] = {'Right PFC HbO': [4, 6, 10, 64],
                         'Right PFC HbR': [5, 7, 11, 65],
                         'Left PFC HbO': [0, 2, 8, 62],
                         'Left PFC HbR': [1, 3, 9, 63]}
        data['event_id'] = {'baseline': 1, 'mental arithmetic': 2}
        data['tmin'] = -2
        data['tmax'] = 10

        # Load data matrices
        cnt = scipy.io.loadmat(f'{path}/{subject}/cnt.mat')
        mrk = scipy.io.loadmat(f'{path}/{subject}/mrk.mat')

        # Get sampling frequency (in Hz)
        data['sfreq'] = cnt['cnt'][0, 1]['fs'][0, 0][0][0]

        # Get original channel info
        wls = cnt['cnt'][0, 1]['wavelengths'][0, 0][0].tolist()
        dpfs = [6., 6.]
        chs = cnt['cnt'][0, 1]['clab'][0, 0][0]
        chs = [str(ch[0]) for ch in chs]
        ch_wls = []
        ch_dpfs = []
        for ch in chs:
            if 'lowWL' in ch:
                ch_wls.append(wls[0])
                ch_dpfs.append(dpfs[0])
            elif 'highWL' in ch:
                ch_wls.append(wls[1])
                ch_dpfs.append(dpfs[1])
        chs = [ch.replace('lowWL', '') for ch in chs]
        chs = [ch.replace('highWL', '') for ch in chs]
        ch_distances = [3 for ch in chs]

        # Load intensities and labels
        ma_idx = [1, 3, 5]
        x_blocks = []
        labels_blocks = []
        for idx in ma_idx:
            x = np.swapaxes(cnt['cnt'][0, idx]['x'][0, 0], 0, 1)
            markers_time = mrk['mrk'][0, idx]['time'][0, 0]/1000*data['sfreq']
            markers_time = np.around(markers_time).astype(int)
            y_idx_ma = np.nonzero(mrk['mrk'][0, idx]['y'][0, 0][0])
            y_idx_bl = np.nonzero(mrk['mrk'][0, idx]['y'][0, 0][1])
            time_ma = markers_time[0, y_idx_ma]
            time_bl = markers_time[0, y_idx_bl]
            labels = np.zeros((1, x.shape[-1]))
            labels[0, time_ma] = 2
            labels[0, time_bl] = 1
            x_blocks.append(x)
            labels_blocks.append(labels)

        # Get channel info and Hb data
        delta_c_blocks = []
        for x_block in x_blocks:
            dod = ns.intensities_to_od_changes(x_block)
            mbll_data = ns.mbll(dod, chs, ch_wls, ch_dpfs, ch_distances,
                                unit='cm', table='wray')
            dc, ch_names, ch_types = mbll_data
            delta_c_blocks.append(dc)
        mne_ch_names = []
        for idx, ch_name in enumerate(ch_names):
            if ch_types[idx] == 'hbo':
                mne_ch_names.append(ch_name + ' HbO')
            elif ch_types[idx] == 'hbr':
                mne_ch_names.append(ch_name + ' HbR')
        data['ch_names'] = mne_ch_names
        data['ch_types'] = ch_types
        delta_c = np.hstack(delta_c_blocks)
        data['delta_c'] = delta_c

        # Get labels
        labels = np.hstack(labels_blocks)
        data['labels'] = labels

        return data


class _DatasetBak2019ME():
    """
    Loader for motor execution data from Bak et al., 2019.
    """

    def __init__(self):
        self.subject_list = ['fNIRS 01', 'fNIRS 02', 'fNIRS 03',
                             'fNIRS 04', 'fNIRS 05', 'fNIRS 06',
                             'fNIRS 07', 'fNIRS 08', 'fNIRS 09',
                             'fNIRS 10', 'fNIRS 11', 'fNIRS 12',
                             'fNIRS 13', 'fNIRS 14', 'fNIRS 15',
                             'fNIRS 16', 'fNIRS 17', 'fNIRS 18',
                             'fNIRS 19', 'fNIRS 20', 'fNIRS 21',
                             'fNIRS 22', 'fNIRS 23', 'fNIRS 24',
                             'fNIRS 25', 'fNIRS 26', 'fNIRS 27',
                             'fNIRS 28', 'fNIRS 29', 'fNIRS 30']

    def load(self, subject, path=PATH_BAK_2019):
        data = {}
        data['sides'] = {'Right MC HbO': range(10, 20),
                         'Right MC HbR': range(30, 40),
                         'Left MC HbO': range(0, 10),
                         'Left MC HbR': range(20, 30)}
        data['event_id'] = {'right': 1, 'left': 2, 'foot': 3}
        data['tmin'] = -1.925
        data['tmax'] = 10

        # Load data matrices
        mat = scipy.io.loadmat(f'{path}/{subject}.mat')

        # Get sampling frequency (in Hz)
        data['sfreq'] = mat['dat']['fs'][0, 0][0, 0]

        # Get channel info
        ch_names = [ch.tolist()[0] for ch in mat['dat']['clab'][0, 0][0]]
        ch_names = [ch.replace('deoxy', ' HbR') for ch in ch_names]
        ch_names = [ch.replace('oxy', ' HbO') for ch in ch_names]
        ch_types = ['hbo' if ch.endswith('O') else 'hbr' for ch in ch_names]
        data['ch_names'] = ch_names
        data['ch_types'] = ch_types

        # Get Hb data
        data_keys = [ch for ch in mat.keys() if ch.startswith('ch')]
        delta_c = np.array([mat[key] for key in data_keys]).squeeze()
        delta_c = delta_c * 1e-3  # convert mM to M
        data['delta_c'] = delta_c

        # Get labels
        markers_time = mat['mrk']['time'][0, 0]/1000*data['sfreq']
        markers_time = np.around(markers_time).astype(int)
        y_idx_right = np.nonzero(mat['mrk']['y'][0, 0][0])
        y_idx_left = np.nonzero(mat['mrk']['y'][0, 0][1])
        y_idx_foot = np.nonzero(mat['mrk']['y'][0, 0][2])
        time_right = markers_time[0, y_idx_right]
        time_left = markers_time[0, y_idx_left]
        time_foot = markers_time[0, y_idx_foot]
        labels = np.zeros((1, delta_c.shape[-1]))
        labels[0, time_right] = 1
        labels[0, time_left] = 2
        labels[0, time_foot] = 3
        data['labels'] = labels

        return data


def load_dataset(dataset, path, bandpass=None, order=4, tddr=False,
                 baseline=(None, 0), roi_sides=False, downsample=10):
    """
    Load and filter one of the open access dataset.

    Parameters
    ----------
    dataset : string
        Dataset to load.
        ``'herff_2014_nb'`` for n-back from Herff et al., 2014
        (epoch interval: -5 to 44 seconds).
        ``'shin_2018_nb'`` for n-back from Shin et al., 2018
        (epoch interval: -2 to 40 seconds).
        ``'shin_2018_wg'`` for word generation from Shin et al., 2018
        (epoch interval: -2 to 10 seconds).
        ``'shin_2016_ma'`` for mental arithmetic from Shin et al., 2016
        (epoch interval: -2 to 10 seconds).
        ``'bak_2019_me'`` for motor execution from Bak et al., 2019
        (epoch interval: -2 to 10 seconds).

    path : string
        Path of the directory of the dataset selected with the ``dataset``
        parameter.

    bandpass : list of floats | None
        Cutoff frequencies of the bandpass Butterworth filter (in Hz). Defaults
        to ``None`` for no filtering.

    order : integer
        Order of the bandpass Butterworth filter.

    tddr : boolean
        Whether to apply temporal derivative distribution repair.

    baseline : None or tuple of length 2
        The time interval to apply baseline correction (in sec). If ``None`` do
        not apply it. If a tuple ``(a, b)`` the interval is between ``a`` and
        ``b`` (in seconds). If ``a`` is ``None`` the beginning of the data is
        used and if ``b`` is ``None`` then ``b`` is set to the end of the
        interval. If ``(None, None)`` all the time interval is used. Correction
        is applied by computing mean of the baseline period and subtracting it
        from the data. The baseline ``(a, b)`` includes both endpoints, i.e.
        all timepoints ``t`` such that ``a <= t <= b``.

    roi_sides : boolean
        Whether to average channels by hemisphere in the task related regions
        of interest.

    downsample : float | None
        Downsample data with the specified frequency (in Hz). Defaults to
        ``10``. Ignored if ``None`` or higher than the dataset's original
        sampling frequency.

    Returns
    -------
    epochs : MNE Epochs object
        MNE epochs data with associated labels. Subject IDs are contained in
        the ``metadata`` property.
    """
    raws = None
    groups = []

    if dataset == 'herff_2014_nb':
        loader = _DatasetHerff2014NB()
    elif dataset == 'shin_2018_nb':
        loader = _DatasetShin2018NB()
    elif dataset == 'shin_2018_wg':
        loader = _DatasetShin2018WG()
    elif dataset == 'shin_2016_ma':
        loader = _DatasetShin2016MA()
    elif dataset == 'bak_2019_me':
        loader = _DatasetBak2019ME()
    else:
        loader = None

    bf = "Loading data: {l_bar}{bar}| Subject {n_fmt}/{total_fmt}"
    disable = False
    if not stdout.isatty():
        try:
            __IPYTHON__
        except NameError:
            disable = True
    pbar = tqdm(
        loader.subject_list, bar_format=bf, disable=disable, ascii=True)
    for subj_id, subj in enumerate(pbar):
        try:
            data = loader.load(subj, path)
        except FileNotFoundError:
            raise FileNotFoundError(
                f"dataset not found, please make sure the dataset has been "
                f"downloaded and the proper path has been provided (cf. "
                f"https://hanbnrd.gitlab.io/benchnirs/install.html)")

        # Create MNE raw object from delta_c
        info = mne.create_info(ch_names=data['ch_names'], sfreq=data['sfreq'],
                               ch_types=data['ch_types'])
        raw = mne.io.RawArray(data['delta_c'], info, verbose=False)

        # TDDR
        if tddr is True:
            raw = _temporal_derivative_distribution_repair(raw)

        # Add labels as a stim channel to the MNE raw object
        info_stim = mne.create_info(['STI'], raw.info['sfreq'], ['stim'])
        stim_raw = mne.io.RawArray(data['labels'], info_stim, verbose=False)
        raw.add_channels([stim_raw], force_update_info=True)

        # Filter raw data
        if bandpass is not None:
            iir_params = dict(order=order, ftype='butter', output='sos')
            raw.filter(*bandpass, method='iir', iir_params=iir_params,
                       verbose=False)

        # Average channels by side
        if roi_sides is True:
            raw = combine_channels(raw, data['sides'], method='mean',
                                   keep_stim=True, verbose=False)

        # Get unlabelled possible segments
        events = mne.find_events(raw, consecutive=True, stim_channel='STI',
                                 shortest_event=1, verbose=False)
        include = list(data['event_id'].values())
        labelleds = mne.pick_events(events, include=include)
        unlabelleds = []
        srange = int((data['tmax'] - data['tmin']) * data['sfreq']) + 1
        s = events[0, 0] + abs(int(data['tmin']*data['sfreq']))
        while s <= events[-1, 0] - int(data['tmax']*data['sfreq']):
            smin = s + int(data['tmin']*data['sfreq'])
            smax = s + int(data['tmax']*data['sfreq'])
            lmins = labelleds[:, 0] + int(data['tmin']*data['sfreq'])
            lmaxs = labelleds[:, 0] + int(data['tmax']*data['sfreq'])
            lmin_overlap = (smin <= lmins) & (lmins < smax)
            lmax_overlap = (smin <= lmaxs) & (lmaxs < smax)
            l_overlap = lmin_overlap | lmax_overlap
            if np.any(l_overlap):
                s = labelleds[l_overlap, 0][-1] + srange
            else:
                unlabelleds.append([s, 0, 999])
                s += srange
        unlabelleds = np.array(unlabelleds)

        # Recreate stim channel with labelled and unlabelled events
        new_events = np.vstack((labelleds, unlabelleds))
        new_events[:, 1] = 0  # make events instantaneous
        raw.add_events(new_events, stim_channel='STI', replace=True)

        # Add subject labels associated with the epochs
        events = mne.find_events(raw, stim_channel='STI', shortest_event=1,
                                 verbose=False)
        data['event_id']['unlabelled'] = 999
        include = list(data['event_id'].values())
        targets = mne.pick_events(events, include=include)
        for _ in targets:
            groups.append(subj)

        # Add to global raw
        if raws is None:
            raws = raw.copy()
        else:
            raws.append(raw.copy())

    # Cut epochs
    all_events = mne.find_events(raws, stim_channel='STI', shortest_event=1,
                                 verbose=False)
    groups = DataFrame(groups, columns=['Subject_ID'])
    epochs = mne.Epochs(raws, all_events, event_id=data['event_id'],
                        metadata=groups, preload=True, tmin=data['tmin'],
                        tmax=data['tmax'], baseline=baseline, verbose=False)

    # Downsample
    if downsample and downsample < epochs.info['sfreq']:
        epochs.resample(downsample)

    return epochs


def load_homer(path, tmin, sfreq):
    """
    Load `yTrials` data from the Homer block average function.

    Parameters
    ----------
    path : string
        Path of the directory containing the `yTrials` files with `.mat` file
        extensions.

    tmin : float
        Start time before the trial onset in seconds (negative or null). Should
        be ``0`` if the trigger onset is on the first time point of the trial
        data.

    sfreq : float
        Sampling frequency in Hz.

    Returns
    -------
    epochs : MNE Epochs object
        MNE epochs data with associated labels. Subject IDs are contained in
        the ``metadata`` property.
    """
    ytrial_files = sorted(glob.glob(f'{path}/*.mat'))
    if not ytrial_files:
        raise FileNotFoundError(
            "no .mat file was found in the specified directory"
        )

    bf = "Loading data: {l_bar}{bar}| Subject {n_fmt}/{total_fmt}"
    disable = False
    if not stdout.isatty():
        try:
            __IPYTHON__
        except NameError:
            disable = True
    pbar = tqdm(ytrial_files, bar_format=bf, disable=disable, ascii=True)
    all_nirs_list = []
    all_labels = []
    all_groups = []
    for file_subj in pbar:
        mat = scipy.io.loadmat(file_subj)
        nirs_list = []
        labels = []
        for i_cond in range(mat['yTrials'].shape[1]):
            yTrials_cond = np.transpose(mat['yTrials'][0, i_cond][0])
            if yTrials_cond.ndim != 4:  # catch case with only one trial
                yTrials_cond = np.expand_dims(yTrials_cond, axis=0)
            hbo_cond = yTrials_cond[:, :, 0, :]
            hbr_cond = yTrials_cond[:, :, 1, :]
            nirs_cond = np.concatenate((hbo_cond, hbr_cond), axis=1)
            nirs_list.append(nirs_cond)
            labels += [i_cond + 1 for _ in range(nirs_cond.shape[0])]
        nirs = np.concatenate(nirs_list, axis=0)
        all_nirs_list.append(nirs)
        all_labels += labels
        all_groups += [Path(file_subj).stem for _ in range(nirs.shape[0])]

    all_nirs = np.concatenate(all_nirs_list, axis=0)
    events = np.zeros((len(all_labels), 3))
    events[:, 2] = all_labels
    events[:, 0] = (
        np.arange(0, len(events)) * nirs.shape[-1] + int(abs(tmin)*sfreq)
    )
    events = events.astype(np.int64)

    ch_names_hbo = [f'{i_ch}_hbo' for i_ch in range(hbo_cond.shape[1])]
    ch_names_hbr = [f'{i_ch}_hbr' for i_ch in range(hbr_cond.shape[1])]
    ch_types_hbo = ['hbo' for _ in range(hbo_cond.shape[1])]
    ch_types_hbr = ['hbr' for _ in range(hbr_cond.shape[1])]
    ch_names = ch_names_hbo + ch_names_hbr
    ch_types = ch_types_hbo + ch_types_hbr

    metadata = DataFrame(all_groups, columns=['Subject_ID'])
    info = mne.create_info(ch_names=ch_names, ch_types=ch_types, sfreq=sfreq)

    if tmin < 0:
        baseline = (None, 0)
    elif tmin == 0:
        baseline = None
    else:
        raise ValueError(f'expected negative or null tmin, got {tmin}')

    epochs = mne.EpochsArray(all_nirs, info, events=events, tmin=tmin,
                             baseline=baseline, metadata=metadata,
                             verbose=False)

    return epochs



import itertools
import matplotlib.pyplot as plt
import numpy as np
import os
import pickle
import seaborn as sns
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from pandas import DataFrame
from torch.utils.data import DataLoader, Dataset
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.metrics import (accuracy_score, precision_recall_fscore_support,
                             confusion_matrix)
from sklearn.model_selection import (KFold, StratifiedKFold, GroupKFold,
                                     GridSearchCV, train_test_split)
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import LinearSVC
from sklearn.utils import shuffle
from sys import stdout
from tqdm import trange


OUTER_K = 5
INNER_K = 3

# Standard machine learning parameters
MAX_ITER = 250000  # for support vector classifier
C_LIST = [1e-3, 1e-2, 1e-1, 1e0]
N_NEIGHBORS_LIST = list(range(1, 10))

# Deep learning parameters
PATIENCE = 5  # for early stopping


def machine_learn(nirs, labels, groups, model, normalize=None,
                  random_state=None, output_folder='./outputs'):
    """
    Perform nested k-fold cross-validation for standard machine learning models
    producing metrics and confusion matrices. The models include linear
    discriminant analysis (LDA), support vector classifier (SVC) with grid
    search for the regularization parameter (inner cross-validation), and
    k-nearest neighbors (kNN) with grid search for the number of neighbors
    (inner cross-validation).

    Parameters
    ----------
    nirs : array of shape (n_samples, n_channels, n_times)
        Processed NIRS data.

    labels : array of integers
        List of labels matching the NIRS data.

    groups : array of integers | None
        List of subject ID matching the NIRS data to perform a group k-fold
        cross-validation. If ``None``, performs a stratified k-fold
        cross-validation instead.

    model : string
        Standard machine learning to use. Either ``'lda'`` for a linear
        discriminant analysis, ``'svc'`` for a linear support vector
        classifier or ``'knn'`` for a k-nearest neighbors classifier.

    normalize : tuple of integers | None
        Axes on which to normalize data before feeding to the model with
        min-max scaling based on the train set for each iteration of the outer
        cross-validation. For example (0, 2) to normalize across samples and
        time. Defaults to ``None`` for no normalization.

    random_state : integer | None
        Controls the shuffling applied to data. Pass an integer for
        reproducible output across multiple function calls. Defaults to
        ``None`` for not setting the seed.

    output_folder : string
        Path to the directory into which the figures will be saved. Defaults to
        ``'./outputs'``.

    Returns
    -------
    accuracies : array of floats
        List of accuracies on the test sets (one for each iteration of the
        outer cross-validation).

    all_hps : list of floats | list of None
        List of regularization parameters for the SVC or a list of None for the
        LDA (one for each iteration of the outer cross-validation).

    additional_metrics : list of tuples
        List of tuples of metrics composed of (precision, recall, F1 score) on
        the outer cross-validation (one tuple for each iteration of the outer
        cross-validation). This uses the ``precision_recall_fscore_support``
        function from scikit-learn with ``average='weighted'``, ``y_true`` and
        ``y_pred`` being the true and the predictions on the specific iteration
        of the outer cross-validation.
    """
    print(f'Machine learning: {model}')

    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    # K-fold cross-validator
    if groups is None:
        out_kf = StratifiedKFold(n_splits=OUTER_K)
        in_kf = StratifiedKFold(n_splits=INNER_K)
    else:
        out_kf = GroupKFold(n_splits=OUTER_K)
        in_kf = GroupKFold(n_splits=INNER_K)
    all_y_true = []
    all_y_pred = []
    accuracies = []
    additional_metrics = []
    all_hps = []
    out_split = out_kf.split(nirs, labels, groups)
    for k, out_idx in enumerate(out_split):
        print(f'    FOLD #{k}')
        nirs_train, nirs_test = nirs[out_idx[0]], nirs[out_idx[1]]
        labels_train, labels_test = labels[out_idx[0]], labels[out_idx[1]]

        if groups is None:
            groups_train = None
            nirs_train, labels_train = shuffle(
                nirs_train, labels_train, random_state=random_state)
        else:
            print(f'    > Test set subject(s): {set(groups[out_idx[1]])}')
            groups_train = groups[out_idx[0]]
            nirs_train, labels_train, groups_train = shuffle(
                nirs_train, labels_train, groups_train,
                random_state=random_state)

        all_y_true += labels_test.tolist()

        # Min-max scaling
        if normalize is not None:
            maxs = nirs_train.max(axis=normalize, keepdims=True)
            mins = nirs_train.min(axis=normalize, keepdims=True)
            nirs_train = (nirs_train - mins) / (maxs - mins)
            nirs_test = (nirs_test - mins) / (maxs - mins)

        nirs_train = nirs_train.reshape(len(nirs_train), -1)
        nirs_test = nirs_test.reshape(len(nirs_test), -1)

        in_split = in_kf.split(nirs_train, labels_train, groups_train)

        # LDA
        if model == 'lda':
            lda = LinearDiscriminantAnalysis()
            lda.fit(nirs_train, labels_train)
            y_pred = lda.predict(nirs_test).tolist()
            all_hps.append(None)

        # SVC
        elif model == 'svc':
            parameters = {'C': C_LIST}
            svc = LinearSVC(max_iter=MAX_ITER, dual='auto')
            clf = GridSearchCV(svc, parameters, scoring='accuracy',
                               cv=in_split)
            clf.fit(nirs_train, labels_train)
            y_pred = clf.predict(nirs_test).tolist()
            all_hps.append(clf.best_params_['C'])

        # kNN
        elif model == 'knn':
            parameters = {'n_neighbors': N_NEIGHBORS_LIST}
            knn = KNeighborsClassifier()
            clf = GridSearchCV(knn, parameters, scoring='accuracy',
                               cv=in_split)
            clf.fit(nirs_train, labels_train)
            y_pred = clf.predict(nirs_test).tolist()
            all_hps.append(clf.best_params_['n_neighbors'])

        # Metrics
        accuracies.append(accuracy_score(labels_test, y_pred))
        prfs = precision_recall_fscore_support(labels_test, y_pred,
                                               average='weighted')
        additional_metrics.append(prfs[:-1])
        all_y_pred += y_pred

    # Figures
    cm = confusion_matrix(all_y_true, all_y_pred, normalize='true')
    sns.heatmap(cm, annot=True, cmap='crest', vmin=0.1, vmax=0.8)
    plt.xlabel('Predicted')
    plt.ylabel('Ground truth')
    plt.title('Confusion matrix on the test sets')
    plt.savefig(f'{output_folder}/confusion_matrix.png')
    plt.close()

    accuracies = np.array(accuracies)

    return accuracies, all_hps, additional_metrics


class _DatasetFromNumPy(Dataset):
    """
    Dataset loader from NumPy arrays for PyTorch.
    """

    def __init__(self, x, y):
        self.x = torch.from_numpy(x)
        self.y = torch.from_numpy(y)

    def __getitem__(self, index):
        return self.x[index], self.y[index]

    def __len__(self):
        return len(self.x)


class _ANNClassifier(nn.Module):

    def __init__(self, n_classes):
        super(_ANNClassifier, self).__init__()
        self.fc1 = nn.Linear(12, 8)
        self.fc2 = nn.Linear(8, 4)
        self.fc3 = nn.Linear(4, n_classes)

    def forward(self, x):
        batch_size = x.size(0)
        x = x.view(batch_size, -1)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


class _CNNClassifier(nn.Module):

    def __init__(self, n_classes):
        super(_CNNClassifier, self).__init__()
        self.conv1 = nn.Conv1d(4, 4, kernel_size=10, stride=2)  # tempo conv
        self.pool1 = nn.MaxPool1d(2)
        self.conv2 = nn.Conv1d(4, 4, kernel_size=5, stride=2)  # tempo conv
        self.pool2 = nn.MaxPool1d(2)
        self.fc1 = nn.Linear(20, 10)
        self.fc2 = nn.Linear(10, n_classes)

    def forward(self, x):
        batch_size = x.size(0)
        x = F.relu(self.conv1(x))
        x = self.pool1(x)
        x = F.relu(self.conv2(x))
        x = self.pool2(x)
        x = x.view(batch_size, -1)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x


class _LSTMClassifier(nn.Module):

    def __init__(self, n_classes):
        super(_LSTMClassifier, self).__init__()
        self.unit_size = 20  # number of timepoints for each elt of the seq
        self.hidden_size = 36
        input_size = self.unit_size * 4  # number of timepoints x 4 channels
        self.lstm = nn.LSTM(input_size, self.hidden_size, batch_first=True)
        self.fc1 = nn.Linear(self.hidden_size, 16)
        self.fc2 = nn.Linear(16, n_classes)

    def forward(self, x):
        # Reshape
        r = x.size(-1) % self.unit_size
        if r > 0:
            x = x[:, :, :-r]  # crop to fit unit size
        x = x.reshape(x.size(0), 4, -1, self.unit_size)  # (b, ch, seq, tpts)
        x = x.permute(0, 2, 1, 3)  # (b, seq, ch, tpts)
        x = x.reshape(x.size(0), x.size(1), -1).double()

        # Initialise hidden and cell states
        h0 = torch.zeros(1, x.size(0), self.hidden_size).double().to(x.device)
        c0 = torch.zeros(1, x.size(0), self.hidden_size).double().to(x.device)

        # Feed to model
        x, _ = self.lstm(x, (h0, c0))
        x = x[:, -1, :]  # last output of the sequence
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x


def _train_dl(nirs_train, labels_train, clf, batch_size, lr, max_epochs,
              earliest_stop, random_state, device):
    """
    Train a deep learning classifier with PyTorch.
    """
    # Load data
    if earliest_stop is not None:
        split = train_test_split(nirs_train, labels_train, shuffle=True,
                                 train_size=0.80, stratify=labels_train,
                                 random_state=random_state)
        nirs_train, labels_train = split[0], split[2]
        nirs_val, labels_val = split[1], split[3]
        dataset_val = _DatasetFromNumPy(nirs_val, labels_val)
        val_loader = DataLoader(dataset=dataset_val, batch_size=1,
                                shuffle=False)
    dataset_train = _DatasetFromNumPy(nirs_train, labels_train)
    train_loader = DataLoader(dataset=dataset_train, batch_size=batch_size,
                              shuffle=True)

    # Instantiate model and hyperparameters
    device_count = torch.cuda.device_count()
    if device_count > 1:
        clf = nn.DataParallel(clf)  # use multiple GPUs
    clf.to(device)
    criterion = nn.CrossEntropyLoss()  # LogSoftmax() & NLLLoss()
    optimizer = optim.Adam(clf.parameters(), lr=lr)
    # scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=25, gamma=0.5)

    train_losses = []
    train_accuracies = []
    val_losses = []
    val_accuracies = []
    bf = "    > Model training: {l_bar}{bar}| Epoch {n_fmt}/{total_fmt}"
    disable = False
    if not stdout.isatty():
        try:
            __IPYTHON__
        except NameError:
            disable = True
    pbar = trange(
        max_epochs, bar_format=bf, leave=False, disable=disable, ascii=True)
    for epoch in pbar:
        # Train
        clf.train()
        running_loss = 0.0
        correct = 0.0
        total = 0.0
        for i, data in enumerate(train_loader):
            # Get the inputs
            x, y = data[0].to(device), data[1].to(device)

            # Zero the parameter gradients
            optimizer.zero_grad()

            # Forward
            outputs = clf(x)
            _, predicted = torch.max(outputs.data, 1)
            loss = criterion(outputs, y)

            # Backward & optimize
            loss.backward()
            optimizer.step()

            # Get statistics
            running_loss += loss.item()
            total += y.size(0)
            correct += (predicted == y).sum()
            correct = int(correct)
        train_losses.append(running_loss / (i+1))
        train_accuracies.append(correct / total)

        # if epoch % 5 == 0:
        #     print(f'Loss: {train_losses[-1]}, '
        #           f'Accuracy: {train_accuracies[-1]}')

        if earliest_stop is not None:
            # Validate
            clf.eval()
            with torch.no_grad():
                running_loss = 0.0
                correct = 0.0
                total = 0.0
                for i, data in enumerate(val_loader):
                    x, y = data[0].to(device), data[1].to(device)
                    outputs = clf(x)
                    _, predicted = torch.max(outputs.data, 1)
                    loss = criterion(outputs, y)
                    running_loss += loss.item()
                    total += y.size(0)
                    correct += (predicted == y).sum()
                    correct = int(correct)
                val_losses.append(running_loss / (i+1))
                val_accuracies.append(correct / total)
                last_sorted = sorted(val_losses[-PATIENCE:])
                if (epoch >= max(earliest_stop, PATIENCE) and
                        val_losses[-PATIENCE:] == last_sorted):
                    print(f'    > Early stopping after {epoch+1} epochs')
                    break
        # scheduler.step()

    if device_count > 1:
        clf = clf.module

    results = {'train_losses': train_losses,
               'train_accuracies': train_accuracies,
               'val_losses': val_losses,
               'val_accuracies': val_accuracies}
    return clf, results


def _test_dl(nirs_test, labels_test, clf, device):
    """
    Test a deep learning classifier with PyTorch.
    """
    # Load data sets
    dataset_test = _DatasetFromNumPy(nirs_test, labels_test)
    test_loader = DataLoader(dataset=dataset_test, batch_size=1,
                             shuffle=False)

    if torch.cuda.device_count() > 1:
        clf = nn.DataParallel(clf)  # use multiple GPUs
    clf.to(device)
    criterion = nn.CrossEntropyLoss()  # LogSoftmax() & NLLLoss()

    # Test
    clf.eval()
    with torch.no_grad():
        y_true = []
        y_pred = []
        correct = 0.0
        total = 0.0
        running_loss = 0.0
        for i, data in enumerate(test_loader):
            x, y = data[0].to(device), data[1].to(device)
            outputs = clf(x)
            _, predicted = torch.max(outputs.data, 1)
            total += y.size(0)
            correct += (predicted == y).sum()
            correct = int(correct)
            y_true.append(y.item())
            y_pred.append(predicted.item())
            loss = criterion(outputs, y)
            running_loss += loss.item()
        results = {'test_loss': running_loss / (i+1),
                   'test_accuracy': correct / total,
                   'y_true': y_true, 'y_pred': y_pred}
    return results


def deep_learn(nirs, labels, groups, model_class, normalize=None,
               batch_sizes=[4, 8, 16, 32, 64],
               lrs=[1e-5, 1e-4, 1e-3, 1e-2, 1e-1],
               max_epochs=100, min_epochs=1,
               random_state=None, output_folder='./outputs'):
    """
    Perform nested k-fold cross-validation for a deep learning model. Produces
    loss graph, accuracy graph and confusion matrice. Early stopping is
    performed (with a validation set of 20 %) after the hyperparameters have
    been selected. The number of classes is deduced from the number of unique
    labels.

    Parameters
    ----------
    nirs : array of shape (n_samples, n_channels, n_times)
        Processed NIRS data.

    labels : array of integers
        List of labels matching the NIRS data.

    groups : array of integers | None
        List of subject ID matching the NIRS data to perform a group k-fold
        cross-validation. If ``None``, performs a stratified k-fold
        cross-validation instead.

    model_class : string | PyTorch nn.Module class
        The PyTorch model class to use. If a string, can be either ``'ann'``,
        ``'cnn'`` or ``'lstm'``. If a PyTorch ``nn.Module`` class, the
        ``__init__()`` method must accept the number of classes as a parameter,
        and this needs to be the number of output neurons.

    normalize : tuple of integers | None
        Axes on which to normalize data before feeding to the model with
        min-max scaling based on the train set for each iteration of the outer
        cross-validation. For example (0, 2) to normalize across samples and
        time. Defaults to ``None`` for no normalization.

    batch_sizes : list of integers
        List of batch sizes to test for fine-tuning.

    lrs : list of floats
        List of learning rates to test for fine-tuning.

    max_epochs : integer
        Maximum number of training epochs possible. Defaults to ``100``.

    min_epochs : integer
        Minimum number of training epochs before early stopping. Defaults to
        ``1``.

    random_state : integer | None
        Controls the shuffling applied to data and random model initialization.
        Pass an integer for reproducible output across multiple function calls.
        Defaults to ``None`` for not setting the seed.

    output_folder : string
        Path to the directory into which the figures will be saved. Defaults to
        ``'./outputs'``.

    Returns
    -------
    accuracies : array of floats
        List of accuracies on the test sets (one for each iteration of the
        outer cross-validation).

    all_hps : list of tuples
        List of best hyperparameters (one tuple for each iteration of the outer
        cross-validation). Each tuple will be `(batch size, learning rate)`.

    additional_metrics : list of tuples
        List of tuples of metrics composed of (precision, recall, F1 score) on
        the outer cross-validation (one tuple for each iteration of the outer
        cross-validation). This uses the ``precision_recall_fscore_support``
        function from scikit-learn with ``average='weighted'``, ``y_true`` and
        ``y_pred`` being the true and the predictions on the specific iteration
        of the outer cross-validation.
    """
    if random_state is not None:
        torch.manual_seed(random_state)

    n_classes = len(set(labels))

    if min_epochs > max_epochs:
        raise ValueError(
            f'expected min_epochs less than or equal to max_epochs, '
            f'got {min_epochs}'
        )

    # Assign PyTorch model class
    if model_class == 'ann':
        model_class = _ANNClassifier
    elif model_class == 'cnn':
        model_class = _CNNClassifier
    elif model_class == 'lstm':
        model_class = _LSTMClassifier

    print(f'Deep learning: {model_class.__name__}', end=' ')

    # Set device
    if torch.cuda.is_available():
        device = torch.device('cuda:0')
        device_count = torch.cuda.device_count()
        print(f"(using {device_count} GPU{'s' if device_count > 1 else ''})")
    else:
        device = torch.device('cpu')
        print("(using CPU)")

    # Outer split
    if os.path.isfile(f'{output_folder}/split.pickle'):
        print('    Saved k-fold split found, loading it...', end=' ')
        with open(f'{output_folder}/split.pickle', 'rb') as f:
            out_split = pickle.load(f)
        print('Done!')
    else:
        if groups is None:
            out_kf = StratifiedKFold(n_splits=OUTER_K)
        else:
            out_kf = GroupKFold(n_splits=OUTER_K)
        out_split = list(out_kf.split(nirs, labels, groups))

        if not os.path.isdir(output_folder):
            os.makedirs(output_folder)
        with open(f'{output_folder}/split.pickle', 'wb') as f:
            pickle.dump(out_split, f)

    # Inner split
    if groups is None:
        in_kf = StratifiedKFold(n_splits=INNER_K)
    else:
        in_kf = GroupKFold(n_splits=INNER_K)

    for k, out_idx in enumerate(out_split):
        print(f'    Training outer fold #{k}')
        nirs_train = nirs[out_idx[0]]
        labels_train = labels[out_idx[0]]

        if groups is None:
            groups_train = None
            nirs_train, labels_train = shuffle(
                nirs_train, labels_train, random_state=random_state)
        else:
            groups_train = groups[out_idx[0]]
            nirs_train, labels_train, groups_train = shuffle(
                nirs_train, labels_train, groups_train,
                random_state=random_state)

        # Min-max scaling
        if normalize is not None:
            maxs = nirs_train.max(axis=normalize, keepdims=True)
            mins = nirs_train.min(axis=normalize, keepdims=True)
            nirs_train = (nirs_train - mins) / (maxs - mins)

        if os.path.isfile(f'{output_folder}/model_k{k}.pt'):
            print('    > Classifier checkpoint found, skipping training')
        else:
            # Train classifier for each combination of hyperparameters
            hp_list = list(itertools.product(batch_sizes, lrs))
            if len(hp_list) > 1:
                print('    > Hyperparameter selection')
                in_accuracies = [[] for _ in hp_list]
                for i, hp in enumerate(hp_list):
                    batch_size, lr = hp[0], hp[1]
                    in_split = in_kf.split(nirs_train, labels_train,
                                           groups_train)
                    for in_idx in in_split:
                        nirs_in_train = nirs_train[in_idx[0]]
                        labels_in_train = labels_train[in_idx[0]]
                        nirs_val = nirs_train[in_idx[1]]
                        labels_val = labels_train[in_idx[1]]

                        clf = model_class(n_classes).double()
                        clf, _ = _train_dl(nirs_in_train, labels_in_train,
                                           clf, batch_size, lr, max_epochs,
                                           None, random_state, device)
                        results = _test_dl(nirs_val, labels_val, clf, device)
                        in_accuracies[i].append(results['test_accuracy'])

                # Get best hyperparameters
                in_average_accuracies = np.mean(in_accuracies, axis=1)
                index_best = np.argmax(in_average_accuracies)
                best_hps = hp_list[index_best]
            else:
                best_hps = (batch_sizes[0], lrs[0])

            # Retrain with best hyperparameters
            clf = model_class(n_classes).double()
            clf, results = _train_dl(nirs_train, labels_train,
                                     clf, best_hps[0], best_hps[1], max_epochs,
                                     min_epochs, random_state, device)

            # Save trained model and training results
            clf.cpu()
            torch.save(clf.state_dict(), f'{output_folder}/model_k{k}.pt')
            if k == 0:
                with open(f'{output_folder}/parameters.txt', 'w') as f:
                    f.write(f'Model\n-----\n{clf}\n\n')
                    f.write('Parameters\n----------\n')
                    f.write(f'normalize = {normalize}\n')
                    f.write(f'batch_sizes = {batch_sizes}\n')
                    f.write(f'lrs = {lrs}\n')
                    f.write(f'max_epochs = {max_epochs}\n')
                    f.write(f'min_epochs = {min_epochs}\n')
                    f.write(f'random_state = {random_state}\n')
            with open(f'{output_folder}/hps_k{k}.pickle', 'wb') as f:
                pickle.dump(best_hps, f)
            with open(f'{output_folder}/results_k{k}.pickle', 'wb') as f:
                pickle.dump(results, f)

            # Plot outer fold loss and accuracy graph
            fig, axes = plt.subplots(ncols=2, figsize=(16, 6))
            title = (f'Selected hyperparameters: batch size = {best_hps[0]}, '
                     f'learning rate = {best_hps[1]}')
            fig.suptitle(title)
            epochs = [epoch for epoch in range(len(results['train_losses']))]
            dict_losses = {'Epoch': epochs,
                           'Training': results['train_losses'],
                           'Validation': results['val_losses']}
            df_losses = DataFrame(dict_losses)
            df_losses = df_losses.melt(
                id_vars=['Epoch'], value_vars=['Training', 'Validation'],
                var_name='Condition', value_name='Loss')
            sns.lineplot(ax=axes[0], data=df_losses, y='Loss',
                         x='Epoch', hue='Condition', estimator=None)
            dict_accuracies = {'Epoch': epochs,
                               'Training': results['train_accuracies'],
                               'Validation': results['val_accuracies']}
            df_accuracies = DataFrame(dict_accuracies)
            df_accuracies = df_accuracies.melt(
                id_vars=['Epoch'], value_vars=['Training', 'Validation'],
                var_name='Condition', value_name='Accuracy')
            sns.lineplot(ax=axes[1], data=df_accuracies, y='Accuracy',
                         x='Epoch', hue='Condition', estimator=None)
            plt.savefig(f'{output_folder}/graph_k{k}.png',
                        bbox_inches='tight')
            plt.close()

    all_ks, all_epochs = [], []
    all_train_losses, all_val_losses = [], []
    all_train_accuracies, all_val_accuracies = [], []
    all_y_true, all_y_pred = [], []
    accuracies, all_hps, additional_metrics = [], [], []
    for k, out_idx in enumerate(out_split):
        print(f'    Testing outer fold #{k}')
        if groups is not None:
            print(f'    > Test set subject(s): {set(groups[out_idx[1]])}')
        nirs_train, nirs_test = nirs[out_idx[0]], nirs[out_idx[1]]
        labels_test = labels[out_idx[1]]

        # Min-max scaling of test set using training set only to avoid leakage
        if normalize is not None:
            maxs = nirs_train.max(axis=normalize, keepdims=True)
            mins = nirs_train.min(axis=normalize, keepdims=True)
            nirs_test = (nirs_test - mins) / (maxs - mins)

        # Load trained model, hyperparameters and training results
        clf = model_class(n_classes).double()
        clf.load_state_dict(torch.load(f'{output_folder}/model_k{k}.pt'))
        with open(f'{output_folder}/hps_k{k}.pickle', 'rb') as f:
            best_hps = pickle.load(f)
        with open(f'{output_folder}/results_k{k}.pickle', 'rb') as f:
            results = pickle.load(f)

        # Append training details
        all_hps.append(best_hps)
        all_ks += [k for _ in results['train_losses']]
        all_epochs += [epoch for epoch in range(len(results['train_losses']))]
        all_train_losses += results['train_losses']
        all_val_losses += results['val_losses']
        all_train_accuracies += results['train_accuracies']
        all_val_accuracies += results['val_accuracies']

        # Test model
        results = _test_dl(nirs_test, labels_test, clf, device)
        all_y_true += results['y_true']
        all_y_pred += results['y_pred']
        accuracies.append(results['test_accuracy'])
        prfs = precision_recall_fscore_support(
            results['y_true'], results['y_pred'], average='weighted')
        additional_metrics.append(prfs[:-1])

    # Plot all loss and accuracy graphs
    _, axes = plt.subplots(ncols=2, figsize=(16, 6))
    dict_losses = {'k': all_ks, 'Epoch': all_epochs,
                   'Training': all_train_losses,
                   'Validation': all_val_losses}
    df_losses = DataFrame(dict_losses)
    df_losses = df_losses.melt(id_vars=['k', 'Epoch'],
                               value_vars=['Training', 'Validation'],
                               var_name='Condition', value_name='Loss')
    sns.lineplot(ax=axes[0], data=df_losses, y='Loss', x='Epoch',
                 hue='Condition', units='k', estimator=None)
    dict_accuracies = {'k': all_ks, 'Epoch': all_epochs,
                       'Training': all_train_accuracies,
                       'Validation': all_val_accuracies}
    df_accuracies = DataFrame(dict_accuracies)
    df_accuracies = df_accuracies.melt(id_vars=['k', 'Epoch'],
                                       value_vars=['Training', 'Validation'],
                                       var_name='Condition',
                                       value_name='Accuracy')
    sns.lineplot(ax=axes[1], data=df_accuracies, y='Accuracy', x='Epoch',
                 hue='Condition', units='k', estimator=None)
    plt.savefig(f'{output_folder}/graphs.png', bbox_inches='tight')
    plt.show()
    plt.close()

    # Figures
    cm = confusion_matrix(all_y_true, all_y_pred, normalize='true')
    sns.heatmap(cm, annot=True, cmap='crest', vmin=0.1, vmax=0.8)
    plt.xlabel('Predicted')
    plt.ylabel('Ground truth')
    plt.title('Confusion matrix on the test sets')
    plt.savefig(f'{output_folder}/confusion_matrix.png')
    plt.close()

    accuracies = np.array(accuracies)

    return accuracies, all_hps, additional_metrics


def _train_encdec(x_train, y_train, encoder, decoder, batch_size, lr,
                  max_epochs, earliest_stop, random_state, device):
    """
    Train a deep learning encoder-decoder with PyTorch.
    """
    # Load data
    if earliest_stop is not None:
        split = train_test_split(x_train, y_train, shuffle=True,
                                 train_size=0.80, random_state=random_state)
        x_train, y_train = split[0], split[2]
        x_val, y_val = split[1], split[3]
        dataset_val = _DatasetFromNumPy(x_val, y_val)
        val_loader = DataLoader(dataset=dataset_val, batch_size=1,
                                shuffle=False)
    dataset_train = _DatasetFromNumPy(x_train, y_train)
    train_loader = DataLoader(dataset=dataset_train, batch_size=batch_size,
                              shuffle=True)

    # Instantiate model and hyperparameters
    device_count = torch.cuda.device_count()
    if device_count > 1:  # use multiple GPUs
        encoder = nn.DataParallel(encoder)
        decoder = nn.DataParallel(decoder)
    encoder.to(device)
    decoder.to(device)
    criterion = nn.MSELoss()
    params = list(encoder.parameters()) + list(decoder.parameters())
    optimizer = optim.Adam(params, lr=lr)

    train_losses = []
    val_losses = []
    bf = "    > Model training: {l_bar}{bar}| Epoch {n_fmt}/{total_fmt}"
    disable = False
    if not stdout.isatty():
        try:
            __IPYTHON__
        except NameError:
            disable = True
    pbar = trange(
        max_epochs, bar_format=bf, leave=False, disable=disable, ascii=True)
    for epoch in pbar:
        # Train
        encoder.train()
        decoder.train()
        running_loss = 0.0
        for i, data in enumerate(train_loader):
            # Get the inputs
            x, y = data[0].to(device), data[1].to(device)

            # Zero the parameter gradients
            optimizer.zero_grad()

            # Forward
            features = encoder(x)
            outputs = decoder(features)
            loss = criterion(outputs, y)

            # Backward & optimize
            loss.backward()
            optimizer.step()

            # Get statistics
            running_loss += loss.item()
        train_losses.append(running_loss / (i+1))

        # if epoch % 5 == 0:
        #     print(f'Loss: {train_losses[-1]}')

        if earliest_stop is not None:
            # Validate
            encoder.eval()
            decoder.eval()
            with torch.no_grad():
                running_loss = 0.0
                for i, data in enumerate(val_loader):
                    x, y = data[0].to(device), data[1].to(device)
                    features = encoder(x)
                    outputs = decoder(features)
                    loss = criterion(outputs, y)
                    running_loss += loss.item()
                val_losses.append(running_loss / (i+1))
                last_sorted = sorted(val_losses[-PATIENCE:])
                if (epoch >= max(earliest_stop, PATIENCE) and
                        val_losses[-PATIENCE:] == last_sorted):
                    print(f'    > Early stopping after {epoch+1} epochs')
                    break

    if device_count > 1:
        encoder = encoder.module
        decoder = decoder.module

    results = {'train_losses': train_losses,
               'val_losses': val_losses}
    return encoder, decoder, results


def _test_encdec(x_test, y_test, encoder, decoder, device):
    """
    Test a deep learning encoder-decoder with PyTorch.
    """
    # Load data sets
    dataset_test = _DatasetFromNumPy(x_test, y_test)
    test_loader = DataLoader(dataset=dataset_test, batch_size=1,
                             shuffle=False)

    if torch.cuda.device_count() > 1:  # use multiple GPUs
        encoder = nn.DataParallel(encoder)
        decoder = nn.DataParallel(decoder)
    encoder.to(device)
    decoder.to(device)
    criterion = nn.MSELoss()

    # Test
    encoder.eval()
    decoder.eval()
    with torch.no_grad():
        running_loss = 0.0
        for i, data in enumerate(test_loader):
            x, y = data[0].to(device), data[1].to(device)
            features = encoder(x)
            outputs = decoder(features)
            loss = criterion(outputs, y)
            running_loss += loss.item()

        results = {'test_loss': running_loss / (i+1)}
    return results


def _proxy_selec(nirs_train, targets_train, groups_train, enc_class, dec_class,
                 batch_sizes, lrs, max_epochs, min_epochs, random_state,
                 output_folder):
    """
    Train and fine-tune encoder-decoder.
    """
    # Set device
    device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

    if random_state is not None:
        torch.manual_seed(random_state)

    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    if groups_train is None:
        in_kf = KFold(n_splits=INNER_K)
    else:
        in_kf = GroupKFold(n_splits=INNER_K)

    # Encoder-decoder fine-tuning
    hp_list = list(itertools.product(batch_sizes, lrs))
    if len(hp_list) > 1:
        print('    > Hyperparameter selection')
        in_losses = [[] for _ in hp_list]
        for i, hp in enumerate(hp_list):
            batch_size, lr = hp[0], hp[1]
            in_split = in_kf.split(nirs_train, targets_train, groups_train)
            for in_idx in in_split:
                nirs_in_train = nirs_train[in_idx[0]]
                targets_in_train = targets_train[in_idx[0]]
                nirs_val = nirs_train[in_idx[1]]
                targets_val = targets_train[in_idx[1]]

                encoder = enc_class().double()
                decoder = dec_class().double()
                encoder, decoder, _ = _train_encdec(
                    nirs_in_train, targets_in_train, encoder, decoder,
                    batch_size, lr, max_epochs, None, random_state, device)

                results = _test_encdec(nirs_val, targets_val, encoder, decoder,
                                       device)
                in_losses[i].append(results['test_loss'])

        # Get best hyperparameters
        in_average_losses = np.mean(in_losses, axis=1)
        index_best = np.argmin(in_average_losses)
        best_hps = hp_list[index_best]
    else:
        best_hps = (batch_sizes[0], lrs[0])

    # Retrain with best hyperparameters
    encoder = enc_class().double()
    decoder = dec_class().double()
    encoder, decoder, results = _train_encdec(
        nirs_train, targets_train, encoder, decoder,
        best_hps[0], best_hps[1], max_epochs, min_epochs, random_state, device)

    # Save trained model and training results
    encoder.cpu()
    decoder.cpu()
    torch.save(encoder.state_dict(), f'{output_folder}/encoder.pt')
    torch.save(decoder.state_dict(), f'{output_folder}/decoder.pt')
    with open(f'{output_folder}/hps.pickle', 'wb') as f:
        pickle.dump(best_hps, f)
    with open(f'{output_folder}/results.pickle', 'wb') as f:
        pickle.dump(results, f)

    # Plot loss graph
    _, ax = plt.subplots(figsize=(12, 6))
    epochs = [epoch for epoch in range(len(results['train_losses']))]
    dict_losses = {'Epoch': epochs,
                   'Training': results['train_losses'],
                   'Validation': results['val_losses']}
    df_losses = DataFrame(dict_losses)
    df_losses = df_losses.melt(id_vars=['Epoch'],
                               value_vars=['Training', 'Validation'],
                               var_name='Condition', value_name='Loss')
    sns.lineplot(ax=ax, data=df_losses, y='Loss', x='Epoch',
                 hue='Condition', estimator=None)
    plt.savefig(f'{output_folder}/graph.png', bbox_inches='tight')
    plt.close()

    return encoder, decoder


def deep_transfer_learn(nirs, labels, groups, enc_class, dec_class,
                        model_class, normalize=None,
                        batch_sizes=[4, 8, 16, 32, 64],
                        lrs=[1e-5, 1e-4, 1e-3, 1e-2, 1e-1],
                        max_epochs=100, min_epochs=1,
                        random_state=None, output_folder='./outputs'):
    """
    Perform nested k-fold cross-validation for a deep transfer learning model.
    Produces loss graphs, accuracy graph and confusion matrice. Early stopping
    is performed (with a validation set of 20 %) after the hyperparameters have
    been selected. The number of classes is deduced from the number of unique
    labels different than ``999``.

    Parameters
    ----------
    nirs : array of shape (n_samples, n_channels, n_times)
        Processed NIRS data.

    labels : array of integers
        List of labels matching the NIRS data. Should include values ``999``
        for the unlabelled samples.

    groups : array of integers | None
        List of subject ID matching the NIRS data to perform a group k-fold
        cross-validation. If ``None``, performs a stratified k-fold
        cross-validation instead.

    enc_class : PyTorch nn.Module class
        The PyTorch encoder class to use for each Hb channel type.

    dec_class : PyTorch nn.Module class
        The PyTorch decoder class to use for each Hb channel type.

    model_class : PyTorch nn.Module class
        The PyTorch classifier class to use for the overall classification
        containing the Hb encoders. The ``__init__()`` method must accept the
        number of classes, an encoder for HbO and an encoder for HbR as
        parameters. The number of classes needs to be the number of output
        neurons.

    normalize : tuple of integers | None
        Axes on which to normalize data before feeding to the model with
        min-max scaling based on the train set for each iteration of the outer
        cross-validation. For example (0, 2) to normalize across samples and
        time. Defaults to ``None`` for no normalization.

    batch_sizes : list of integers
        List of batch sizes to test for fine-tuning.

    lrs : list of floats
        List of learning rates to test for fine-tuning.

    max_epochs : integer
        Maximum number of training epochs possible. Defaults to ``100``.

    min_epochs : integer
        Minimum number of training epochs before early stopping. Defaults to
        ``1``.

    random_state : integer | None
        Controls the shuffling applied to data and random model initialization.
        Pass an integer for reproducible output across multiple function calls.
        Defaults to ``None`` for not setting the seed.

    output_folder : string
        Path to the directory into which the figures will be saved. Defaults to
        ``'./outputs'``.

    Returns
    -------
    accuracies : array of floats
        List of accuracies for the overall classifier on the test sets (one for
        each iteration of the outer cross-validation).

    all_hps : list of tuples
        List of best hyperparameters for the overall classifier (one tuple for
        each iteration of the outer cross-validation). Each tuple will be
        `(batch size, learning rate)`.

    additional_metrics : list of tuples
        List of tuples of metrics composed of (precision, recall, F1 score) on
        the outer cross-validation (one tuple for each iteration of the outer
        cross-validation). This uses the ``precision_recall_fscore_support``
        function from scikit-learn with ``average='weighted'``, ``y_true`` and
        ``y_pred`` being the true and the predictions on the specific iteration
        of the outer cross-validation.

    :meta private:
    """
    if random_state is not None:
        torch.manual_seed(random_state)

    print(f'Labels: {set(labels)}')
    if 999 in set(labels):
        n_classes = len(set(labels)) - 1  # minus unlabelled
    else:
        n_classes = len(set(labels))

    if min_epochs > max_epochs:
        raise ValueError(
            f'expected min_epochs less than or equal to max_epochs, '
            f'got {min_epochs}'
        )

    print(f'Deep transfer learning: {enc_class.__name__}/'
          f'{dec_class.__name__}-{model_class.__name__}', end=' ')

    # Set device
    if torch.cuda.is_available():
        device = torch.device('cuda:0')
        device_count = torch.cuda.device_count()
        print(f"(using {device_count} GPU{'s' if device_count > 1 else ''})")
    else:
        device = torch.device('cpu')
        print("(using CPU)")

    # Get index to split channel types
    mid_idx = nirs.shape[1] / 2
    if mid_idx.is_integer():
        mid_idx = int(mid_idx)

    # Outer split
    if os.path.isfile(f'{output_folder}/split.pickle'):
        print('    Saved k-fold split found, loading it...', end=' ')
        with open(f'{output_folder}/split.pickle', 'rb') as f:
            out_split = pickle.load(f)
        print('Done!')
    else:
        if groups is None:
            out_kf = StratifiedKFold(n_splits=OUTER_K)
        else:
            out_kf = GroupKFold(n_splits=OUTER_K)
        out_split = list(out_kf.split(nirs, labels, groups))

        if not os.path.isdir(output_folder):
            os.makedirs(output_folder)
        with open(f'{output_folder}/split.pickle', 'wb') as f:
            pickle.dump(out_split, f)

    # Inner split
    if groups is None:
        in_kf = StratifiedKFold(n_splits=INNER_K)
    else:
        in_kf = GroupKFold(n_splits=INNER_K)

    for k, out_idx in enumerate(out_split):
        print(f'    Training outer fold #{k}')
        nirs_train = nirs[out_idx[0]]
        labels_train = labels[out_idx[0]]

        if groups is None:
            groups_train = None
            nirs_train, labels_train = shuffle(
                nirs_train, labels_train, random_state=random_state)
        else:
            groups_train = groups[out_idx[0]]
            nirs_train, labels_train, groups_train = shuffle(
                nirs_train, labels_train, groups_train,
                random_state=random_state)

        # Min-max scaling
        if normalize is not None:
            maxs = nirs_train.max(axis=normalize, keepdims=True)
            mins = nirs_train.min(axis=normalize, keepdims=True)
            nirs_train = (nirs_train - mins) / (maxs - mins)

        # Train and optimise self-supervised models
        nirs_train_hbo = nirs_train[:, mid_idx:, :]
        nirs_train_hbr = nirs_train[:, :mid_idx, :]
        # HbO -> HbR
        if os.path.isfile(f'{output_folder}/k{k}/hbo/encoder.pt'):
            print('    > HbO encoder checkpoint found, loading it...', end=' ')
            enc_hbo = enc_class().double()
            enc_hbo.load_state_dict(
                torch.load(f'{output_folder}/k{k}/hbo/encoder.pt'))
            print('Done!')
        else:
            enc_hbo, _ = _proxy_selec(
                nirs_train_hbo, nirs_train_hbr, groups_train,
                enc_class, dec_class, batch_sizes=batch_sizes, lrs=lrs,
                max_epochs=max_epochs, min_epochs=min_epochs,
                random_state=random_state,
                output_folder=f'{output_folder}/k{k}/hbo')
        # HbR -> HbO
        if os.path.isfile(f'{output_folder}/k{k}/hbr/encoder.pt'):
            print('    > HbR encoder checkpoint found, loading it...', end=' ')
            enc_hbr = enc_class().double()
            enc_hbr.load_state_dict(
                torch.load(f'{output_folder}/k{k}/hbr/encoder.pt'))
            print('Done!')
        else:
            enc_hbr, _ = _proxy_selec(
                nirs_train_hbr, nirs_train_hbo, groups_train,
                enc_class, dec_class, batch_sizes=batch_sizes, lrs=lrs,
                max_epochs=max_epochs, min_epochs=min_epochs,
                random_state=random_state,
                output_folder=f'{output_folder}/k{k}/hbr')

        if os.path.isfile(f'{output_folder}/k{k}/clf.pt'):
            print('    > Classifier checkpoint found, skipping training')
        else:
            # Train classifier for each combination of hyperparameters
            hp_list = list(itertools.product(batch_sizes, lrs))
            if len(hp_list) > 1:
                print('    > Hyperparameter selection')
                in_accuracies = [[] for _ in hp_list]
                for i, hp in enumerate(hp_list):
                    batch_size, lr = hp[0], hp[1]
                    in_split = in_kf.split(nirs_train, labels_train,
                                           groups_train)
                    for in_idx in in_split:
                        nirs_in_train = nirs_train[in_idx[0]]
                        labels_in_train = labels_train[in_idx[0]]
                        nirs_val = nirs_train[in_idx[1]]
                        labels_val = labels_train[in_idx[1]]

                        # Remove unlabelled examples
                        idx_in_train = np.where(
                            np.array(labels_in_train) != 999)
                        labels_in_train = labels_in_train[idx_in_train]
                        nirs_in_train = nirs_in_train[idx_in_train]
                        idx_val = np.where(np.array(labels_val) != 999)
                        labels_val = labels_val[idx_val]
                        nirs_val = nirs_val[idx_val]

                        clf = model_class(n_classes, enc_hbo, enc_hbr).double()
                        clf, _ = _train_dl(nirs_in_train, labels_in_train,
                                           clf, batch_size, lr, max_epochs,
                                           None, random_state, device)
                        results = _test_dl(nirs_val, labels_val, clf, device)
                        in_accuracies[i].append(results['test_accuracy'])

                # Get best hyperparameters
                in_average_accuracies = np.mean(in_accuracies, axis=1)
                index_max = np.argmax(in_average_accuracies)
                best_hps = hp_list[index_max]
            else:
                best_hps = (batch_sizes[0], lrs[0])

            # Retrain with best hyperparameters
            idx_train = np.where(np.array(labels_train) != 999)
            labels_train = labels_train[idx_train]
            nirs_train = nirs_train[idx_train]
            clf = model_class(n_classes, enc_hbo, enc_hbr).double()
            clf, results = _train_dl(nirs_train, labels_train,
                                     clf, best_hps[0], best_hps[1], max_epochs,
                                     min_epochs, random_state, device)

            # Save trained model and training results
            clf.cpu()
            torch.save(clf.state_dict(), f'{output_folder}/k{k}/clf.pt')
            if k == 0:
                with open(f'{output_folder}/parameters.txt', 'w') as f:
                    f.write(f'Models\n------\n{clf}\n{enc_hbo}\n{enc_hbr}\n\n')
                    f.write('Parameters\n----------\n')
                    f.write(f'normalize = {normalize}\n')
                    f.write(f'batch_sizes = {batch_sizes}\n')
                    f.write(f'lrs = {lrs}\n')
                    f.write(f'max_epochs = {max_epochs}\n')
                    f.write(f'min_epochs = {min_epochs}\n')
                    f.write(f'random_state = {random_state}\n')
            with open(f'{output_folder}/k{k}/hps.pickle', 'wb') as f:
                pickle.dump(best_hps, f)
            with open(f'{output_folder}/k{k}/results.pickle', 'wb') as f:
                pickle.dump(results, f)

            # Plot outer fold loss graph
            fig, ax = plt.subplots(figsize=(12, 6))
            title = (f'Selected hyperparameters: batch size = {best_hps[0]}, '
                     f'learning rate = {best_hps[1]}')
            fig.suptitle(title)
            epochs = [epoch for epoch in range(len(results['train_losses']))]
            dict_losses = {'Epoch': epochs,
                           'Training': results['train_losses'],
                           'Validation': results['val_losses']}
            df_losses = DataFrame(dict_losses)
            df_losses = df_losses.melt(
                id_vars=['Epoch'], value_vars=['Training', 'Validation'],
                var_name='Condition', value_name='Loss')
            sns.lineplot(ax=ax, data=df_losses, y='Loss', x='Epoch',
                         hue='Condition', estimator=None)
            plt.savefig(f'{output_folder}/k{k}/clf_graph.png',
                        bbox_inches='tight')
            plt.close()

    all_ks, all_epochs = [], []
    all_train_losses, all_val_losses = [], []
    all_train_accuracies, all_val_accuracies = [], []
    all_y_true, all_y_pred = [], []
    accuracies, all_hps, additional_metrics = [], [], []
    for k, out_idx in enumerate(out_split):
        print(f'    Testing outer fold #{k}')
        if groups is not None:
            print(f'    > Test set subject(s): {set(groups[out_idx[1]])}')
        nirs_train, nirs_test = nirs[out_idx[0]], nirs[out_idx[1]]
        labels_test = labels[out_idx[1]]

        # Min-max scaling of test set using training set only to avoid leakage
        if normalize is not None:
            maxs = nirs_train.max(axis=normalize, keepdims=True)
            mins = nirs_train.min(axis=normalize, keepdims=True)
            nirs_test = (nirs_test - mins) / (maxs - mins)

        # Load trained model, hyperparameters and training results
        enc_hbo = enc_class().double()
        enc_hbr = enc_class().double()
        enc_hbo.load_state_dict(
            torch.load(f'{output_folder}/k{k}/hbo/encoder.pt'))
        enc_hbr.load_state_dict(
            torch.load(f'{output_folder}/k{k}/hbr/encoder.pt'))
        clf = model_class(n_classes, enc_hbo, enc_hbr).double()
        clf.load_state_dict(torch.load(f'{output_folder}/k{k}/clf.pt'))
        with open(f'{output_folder}/k{k}/hps.pickle', 'rb') as f:
            best_hps = pickle.load(f)
        with open(f'{output_folder}/k{k}/results.pickle', 'rb') as f:
            results = pickle.load(f)

        # Append training details
        all_hps.append(best_hps)
        all_ks += [k for _ in results['train_losses']]
        all_epochs += [epoch for epoch in range(len(results['train_losses']))]
        all_train_losses += results['train_losses']
        all_val_losses += results['val_losses']
        all_train_accuracies += results['train_accuracies']
        all_val_accuracies += results['val_accuracies']

        # Test model
        idx_test = np.where(np.array(labels_test) != 999)
        labels_test = labels_test[idx_test]
        nirs_test = nirs_test[idx_test]
        results = _test_dl(nirs_test, labels_test, clf, device)
        all_y_true += results['y_true']
        all_y_pred += results['y_pred']
        accuracies.append(results['test_accuracy'])
        prfs = precision_recall_fscore_support(
            results['y_true'], results['y_pred'], average='weighted')
        additional_metrics.append(prfs[:-1])

    # Plot all loss and accuracy graphs
    _, axes = plt.subplots(ncols=2, figsize=(16, 6))
    dict_losses = {'k': all_ks, 'Epoch': all_epochs,
                   'Training': all_train_losses,
                   'Validation': all_val_losses}
    df_losses = DataFrame(dict_losses)
    df_losses = df_losses.melt(id_vars=['k', 'Epoch'],
                               value_vars=['Training', 'Validation'],
                               var_name='Condition', value_name='Loss')
    sns.lineplot(ax=axes[0], data=df_losses, y='Loss', x='Epoch',
                 hue='Condition', units='k', estimator=None)
    dict_accuracies = {'k': all_ks, 'Epoch': all_epochs,
                       'Training': all_train_accuracies,
                       'Validation': all_val_accuracies}
    df_accuracies = DataFrame(dict_accuracies)
    df_accuracies = df_accuracies.melt(id_vars=['k', 'Epoch'],
                                       value_vars=['Training', 'Validation'],
                                       var_name='Condition',
                                       value_name='Accuracy')
    sns.lineplot(ax=axes[1], data=df_accuracies, y='Accuracy', x='Epoch',
                 hue='Condition', units='k', estimator=None)
    plt.savefig(f'{output_folder}/graphs.png', bbox_inches='tight')
    plt.show()
    plt.close()

    # Figures
    cm = confusion_matrix(all_y_true, all_y_pred, normalize='true')
    sns.heatmap(cm, annot=True, cmap='crest', vmin=0.1, vmax=0.8)
    plt.xlabel('Predicted')
    plt.ylabel('Ground truth')
    plt.title('Confusion matrix on the test sets')
    plt.savefig(f'{output_folder}/confusion_matrix.png')
    plt.close()

    accuracies = np.array(accuracies)

    return accuracies, all_hps, additional_metrics


def train_final(nirs, labels, model_class, batch_size, lr, n_epochs,
                normalize=None, random_state=None, output_folder='./'):
    """
    Train a final neural network classifier on the whole data with the selected
    hyperparameters. The trained neural network checkpoint is saved in the
    output folder. The number of classes is deduced from the number of unique
    labels.

    Parameters
    ----------
    nirs : array of shape (n_samples, n_channels, n_times)
        Processed NIRS data.

    labels : array of integers
        List of labels matching the NIRS data.

    model_class : PyTorch nn.Module class
        The PyTorch model class to use. The ``__init__()`` method must accept
        the number of classes as a parameter, and this needs to be the number
        of output neurons.

    batch_size : integer
        Number of samples per batch.

    lr : float
        Learning rate for the optimizer.

    n_epochs : integer
        Number of training epochs (number of passes over the whole dataset).

    normalize : tuple of integers | None
        Axes on which to normalize data before feeding to the model with
        min-max scaling. For example (0, 2) to normalize across samples and
        time. Defaults to ``None`` for no normalization.

    random_state : integer | None
        Controls the shuffling applied to data and random model initialization.
        Pass an integer for reproducible output across multiple function calls.
        Defaults to ``None`` for not setting the seed.

    output_folder : string
        Path to the directory into which the checkpoint and training graph will
        be saved. Defaults to the current directory.

    Returns
    -------
    clf : PyTorch nn.Module
        The trained PyTorch neural network.
    """
    if random_state is not None:
        torch.manual_seed(random_state)

    # Create output folder if necessary
    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    # Instantiate the model
    n_classes = len(set(labels))
    clf = model_class(n_classes).double()

    print(f'Final training: {model_class.__name__}', end=' ')

    # Set device
    device_count = torch.cuda.device_count()
    if torch.cuda.is_available():
        device = torch.device('cuda:0')
        print(
            f"(using {device_count} GPU{'s' if device_count > 1 else ''})"
        )
    else:
        device = torch.device('cpu')
        print("(using CPU)")

    if os.path.isfile(f'{output_folder}/final_model.pt'):
        print('Checkpoint found, loading it...', end=' ')
        clf.load_state_dict(torch.load(f'{output_folder}/final_model.pt'))
        print('Done!')
    else:
        # Min-max scaling
        if normalize is not None:
            maxs = nirs.max(axis=normalize, keepdims=True)
            mins = nirs.min(axis=normalize, keepdims=True)
            nirs = (nirs - mins) / (maxs - mins)

        # Load data
        dataset_train = _DatasetFromNumPy(nirs, labels)
        train_loader = DataLoader(dataset=dataset_train, batch_size=batch_size,
                                  shuffle=True)

        # Prepare for training
        if device_count > 1:
            clf = nn.DataParallel(clf)  # use multiple GPUs
        clf.to(device)
        criterion = nn.CrossEntropyLoss()  # LogSoftmax() & NLLLoss()
        optimizer = optim.Adam(clf.parameters(), lr=lr)
        # scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=25,
        #                                       gamma=0.5)

        # Training loop
        train_losses = []
        bf = "Model training: {l_bar}{bar}| Epoch {n_fmt}/{total_fmt}"
        disable = False
        if not stdout.isatty():
            try:
                __IPYTHON__
            except NameError:
                disable = True
        pbar = trange(n_epochs, bar_format=bf, disable=disable, ascii=True)
        clf.train()
        for epoch in pbar:
            # Train
            running_loss = 0.0
            for i, data in enumerate(train_loader):
                # Get the inputs
                x, y = data[0].to(device), data[1].to(device)

                # Zero the parameter gradients
                optimizer.zero_grad()

                # Forward
                outputs = clf(x)
                loss = criterion(outputs, y)

                # Backward & optimize
                loss.backward()
                optimizer.step()

                # Get statistics
                running_loss += loss.item()
            train_losses.append(running_loss / (i+1))
            # scheduler.step()

        if device_count > 1:
            clf = clf.module

        # Save trained model
        clf.eval()
        clf.cpu()
        torch.save(clf.state_dict(), f'{output_folder}/final_model.pt')
        with open(f'{output_folder}/final_parameters.txt', 'w') as f:
            f.write(f'Model\n-----\n{clf}\n\n')
            f.write('Parameters\n----------\n')
            f.write(f'batch_size = {batch_size}\n')
            f.write(f'lr = {lr}\n')
            f.write(f'n_epochs = {n_epochs}\n')
            f.write(f'normalize = {normalize}\n')
            f.write(f'random_state = {random_state}\n')

        # Plot training loss
        _, ax = plt.subplots(figsize=(12, 6))
        epochs = [epoch for epoch in range(len(train_losses))]  
        df_losses = DataFrame({'Epoch': epochs, 'Training loss': train_losses})
        sns.lineplot(ax=ax, data=df_losses, y='Training loss', x='Epoch')
        plt.savefig(f'{output_folder}/final_graph.png', bbox_inches='tight')
        plt.show()
        plt.close()

    return clf



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="benchnirs",
    version="1.4.4",
    author="Johann Benerradi",
    author_email="johann.benerradi@gmail.com",
    description="Benchmarking framework for machine learning with fNIRS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/HanBnrd/benchnirs",
    license="GNU GPLv3+",
    packages=setuptools.find_packages(),
    install_requires=[
        "lazy_loader",
        "tqdm",
        "numpy<2",
        "pandas",
        "scipy",
        "mne",
        "matplotlib",
        "seaborn",
        "scikit-learn",
        "torch",
        "nirsimple"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)
